/* Mpeg decoding */
#include "demuxer.hh"
#include "encoder.hh"
#include "a52dec/a52.h"
/* SPU */
#include "../libspu/spu.h"
/* Other */
#include "plugin-loader.hh"
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <math.h>
#include <pthread.h>
#include "fast_memcpy.hh"
#include "../src/external.h"


/* Defines */
#define DEMUX_BUFFER_SIZE 1048576       // 1Mb
#define AUDIO_BUFFER_SIZE 30  *44100*4  // 30 seconds buffer, 5Mb
#define PTSMAX 8192

// loggin
//#define DEBUG
#ifdef DEBUG
#warning         -------------------------------------
#warning
#warning
#warning
#warning
#warning
#warning         DEBUGGING WILL PRODUCE HUGE LOG FILES
#warning
#warning
#warning
#warning
#warning
#warning         -------------------------------------
#endif

// writeout audio channel 0 to /tmp/writeout_audio.raw 
//#define WRITEOUT_AUDIO


/* Globals */
static mpeg2dec_t mpeg2dec;
//static vo_open_t * video_output_open = NULL;
static ao_open_t *audio_output_open[2];
static vo_instance_t *video_output;
//static vo_open_t *video_driver = NULL;
static ao_instance_t *audio_output[2];
//static guint8 demux_buffer[DEMUX_BUFFER_SIZE];
static mpeg_read_buffer_t read_buffer[2];
static gboolean mpeg2_reader_handler_active = FALSE;
static gint mpeg2_reader_handler_pid;
static FILE *in_file;
static gint width = -1,height=-1;
static a52_state_t *output_liba52[2];
static glong total_frames = 0;
static gdouble framerate;
gboolean demuxing = FALSE;
//static gboolean demuxing_frames = FALSE;
static gboolean process = FALSE;
static void(*processor)(guint8*[3],uint32_t pts);
static gint16 gint16_samples[AUDIO_BUFFER_SIZE]; //6*256*6*sizeof(gint16)];
static gint16 *audio_buffer[2];
static gint audio_buffer_index[2];
static gint audio_buffer_base[2];
static gint audio_buffer_total[2];
static gint audio_fetched_total[2];
static gint audio_sample_rate,audio_bit_rate,audio_channels;
//static int demux_pid = 0;
static gint mpeg2_reader_id;
static pthread_t mpeg2_reader;
static pthread_mutex_t mutexmpeg2;
glong SCRstream, SCRvideo, PTSprivate;
glong SCRstream_prev = 0;
static glong IDnumber[2];
static gint PTSindex[2];
gint PTSvideo = -1;
gint PTSaudio = -1;
gdouble PTSframe; // Here PTS value of latest video frame is stored.. or should be, not 100% garanteed for now, couse libmpeg2 cant supply the correct value
static gint16 silence[512];
static gint16 *audio_buffer_chunk;//[2048];
//static gint16 audio_buffer_stable[32768];
static gdouble AUDIO_BYTES_PER_FRAME = -1;//drip_audio_samplerate() / drip_mpeg_framerate() * 4;
static gdouble PTS_PER_AUDIO_FRAME = 480;
static gdouble PTS_PER_AUDIO_BYTE = PTS_PER_AUDIO_FRAME/1024;
static gboolean non_silence[2];
common_instance_t *vo_open_instance;
gchar **input_files;
gint input_files_index = 0;
gint spu_track;
gint audio_track;
glong avifile_audio_outstream_bytes = 0;
gdouble avifile_audio_outstream_bytes_calculated = 0;
/* A52 vari's */
static guint8 *a52_buffer[2];
static guint8 *bufptr[2];
static guint8 *bufpos[2];
static gdouble PTSa52[2];
static guint8* start[2];
static gint avifile_audio_total_offset[2];
static gboolean offset_corrected[2];
static gboolean decoding_audio = FALSE;
        #ifdef WRITEOUT_AUDIO
        gint writeout_audio_fd;
        #endif




/* ----------- Functions -- */

/* Return audio samplerate */
gint drip_audio_samplerate(void) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"drip_audio_samplerate = %i",audio_sample_rate);
    switch (audio_sample_rate) {
        case 0:  return 48000;
                 break;
        default: return audio_sample_rate;
                 break;
    }
}


/* Return audio bitrate */
gint drip_audio_bitrate(void) {
    return audio_bit_rate;
}


/* Return audio bitrate */
gint drip_audio_channels(void) {
    return audio_channels;
}


/* Return width of current open mpeg */
gint drip_mpeg_width(void) {
    return width;
}


/* Return height of current open mpeg */
gint drip_mpeg_height(void) {
    return height;
}


/* Determine framerate out frame_rate_code */
inline void drip_framerate(mpeg2dec_t* _mpeg2dec) {
    switch (_mpeg2dec->frame_rate_code) {
    case 1: framerate = 23.976;
            break;
    case 2: framerate = 24;
            break;
    case 3: framerate = 25;
            break;
    case 4: framerate = 29.97;
    #ifdef DEBUG 
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"drip_framerate, framerate=%f (%f)",framerate,29.97);
    #endif

            break;
    case 5: framerate = 30;
            break;
    case 6: framerate = 50;
            break;
    case 7: framerate = 59.94;
            break;
    case 8: framerate = 60;
            break;
    default:framerate = -1;
            break;
    }
}


/* Return framerate of current open mpeg */
gdouble drip_mpeg_framerate(void) {
    return framerate;
}


/* Return TimeStamp value per video frame */
gdouble TimeStamp_Frame(void) {
    static gdouble TF;
    TF = 90000/framerate;
    #ifdef DEBUG 
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"TimeStamp_Frame: TF=%f  (FR=%f)",TF,framerate);
    #endif
    return TF;
}


/* Seek to percentage in mpeg file */
gint drip_mpeg_seek_percentage(gdouble perc) {
    gint result,pos;
    /* Check perc */
    if (perc<0 || perc>1) {
        perc = 0;
    }
    /* Calc position in bytes from perc */
    result = fseek(in_file,0,SEEK_END);
    pos = (gint)(ftell(in_file) * perc);
    /* Seek to pos */
    result = fseek(in_file,pos,SEEK_SET);
    return result;
}


/* float_to_int */
static inline gint16 convert(int32_t i) {
    if (i>0x43c07fff)
        return 32767;
    else if (i<0x43bf8000)
        return -32768;
    else
        return i-0x43c00000;
}


/* Typecast floats to int16s */
static inline gint float_to_int_stereo(float *_f, gint16 *s16) {
    guint i;
    int32_t *f = (int32_t*)_f;       /* XXX assumes IEEE float format */
    for (i=0; i<256; i++) {
        s16[i*2] = convert(f[i]);
        s16[i*2+1] = convert(f[i+256]);
    }
    return 1024;//256*2*(sizeof(gint16));
}


/* Reset audio buffer */
inline void reset_audio_buffer(gint channel) {
    audio_buffer_index[channel] = 0;
    PTSindex[channel] = 0;
    non_silence[channel] = FALSE;
    //PTSbase[channel] = 0;
    return;
}


/* Debug audio buffer, display status */
inline void audio_buffer_status(gint channel) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"index=%i    base=%i   fetched=%i    stored=%i    bufferd=%i",audio_buffer_index[channel],audio_buffer_base[channel],audio_fetched_total[channel],audio_buffer_total[channel],audio_buffer_total[channel]-audio_fetched_total[channel]);
}


/* Returns audio buffer and resets index */
gint16* fetch_from_audio_buffer(gint samples, gint SCR,gint channel) {
    gint chunk_size = samples; // TODO: 2->channels

    /* Anything in buffer? */
    if ((audio_buffer_index[channel]==0)/* || (PTSindex<1)*/) {
        #ifdef DEBUG 
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio Fetch buffer empty (%i audio frames)",chunk_size,PTSindex);
        #endif
        return NULL;
    }

    /* Enough in buffer? */
    if (audio_buffer_index[channel]<chunk_size) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Fetch buffer holds %ibytes, could not release %ibytes",audio_buffer_index[channel],chunk_size);
        #endif
        return NULL;
    }
    /* Moronic chuck_size? */
    if (chunk_size<1) {
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio Fetch buffer request for %ib canceled",chunk_size);
        return NULL;
    }

    /* Keep track total bytes put here */
    audio_fetched_total[channel]+=chunk_size;

    /* Update buffer */
    // TODO add mutex locking here and in add_to_audio_buffer(..)
    audio_buffer_index[channel] -= chunk_size;
    if(audio_buffer_index[channel]<0) {
        audio_fetched_total[channel] += audio_buffer_index[channel];
        audio_buffer_index[channel]=0;
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio Fetch buffer underflow");
        #endif
    }
    #ifdef DEBUG
    audio_buffer_status(channel);
    #endif

    /* Copy audio data into return pointer */
    audio_buffer_chunk = (gint16*)((glong)audio_buffer[channel]+(glong)audio_buffer_base[channel]);
    /* Update buffer base (offset) */
    audio_buffer_base[channel] += chunk_size;

    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Fetch buffer returned chunk");
    #endif
    return audio_buffer_chunk;
}


/* Builds audio buffer for encoder */
void add_to_audio_buffer(gint16* audio_chunk, gint audio_chunk_size, gint PTS, gint channel) {
    static gint16* audio_buffer_dest;
    static gint i;
    static gdouble PTSoff[2];
    static gdouble PTSstored[2];
    static guint8** chunk = NULL;
    glong audio_chunk_extra;

    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Add buffer new sample, channel %i, PTSstored = %g,   PTS = %i,   sample bytes = %i",channel,PTSstored[channel],PTS,audio_chunk_size);
    #endif

    /* alloc chunk */
    if (chunk == NULL) {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Add buffer new sample, allocated chunk");
        chunk = (guint8**)malloc(3*sizeof(guint8*));
        for (i=0;i<AMAX;i++) {
            PTSoff[i] = 0;
            PTSstored[i] = 0;
        }
        audio_buffer_chunk = (gint16*)malloc(2048);
    }

    /* Process audio data by the filters */
    chunk[0] = (guint8*)audio_chunk;
    chunk[1] = (guint8*)NULL;
    chunk[2] = (guint8*)NULL;
    /* - Plugin Filters, type AUDIO PRE - */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==AUDIO && filters[i].module_phase==PRE) {
            chunk = (*filters[i].module_apply_function)(chunk,audio_chunk_size,(gulong)SCRframe,PTS);
        }
    }
    /* - Plugin Filters, type AUDIO NORMAL - */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==AUDIO && filters[i].module_phase==NORMAL) {
            chunk = (*filters[i].module_apply_function)(chunk,audio_chunk_size,(gulong)SCRframe,PTS);
        }
    }
    /* - Plugin Filters, type AUDIO POST - */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==AUDIO && filters[i].module_phase==POST) {
            chunk = (*filters[i].module_apply_function)(chunk,audio_chunk_size,(gulong)SCRframe,PTS);
        }
    }

    /* Did a plugin allocated new memory? */
    if (audio_chunk != (gint16*)chunk[0]) {
        free(audio_chunk);
        audio_chunk = (gint16*)chunk[0];
    }

    /* Less stored as needed? */
    if (PTS>PTSstored[channel]) {
        /* Add silence chunk */
        audio_chunk_extra = (glong)(((gdouble)PTS-PTSstored[channel])/PTS_PER_AUDIO_BYTE);
        audio_chunk_extra = ((glong)(audio_chunk_extra/4))*4; // round to 4
        if(audio_chunk_extra>0) {
            if (audio_chunk_extra<AUDIO_BUFFER_SIZE) {
                memmove((gint16*)((glong)audio_chunk+audio_chunk_extra),audio_chunk,audio_chunk_size);
                memset(audio_chunk,0,audio_chunk_extra);
                audio_chunk_size += audio_chunk_extra;
                g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Add buffer added %li extra bytes",audio_chunk_extra);
            } else {
                g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio Add buffer couldn't add silence, too large (%libytes)",audio_chunk_extra);
            }
        }
    } 

    /* Track stored audio */
    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Add buffer, PTSstored = %f, new PTSstored = %f",PTSstored[channel],PTSstored[channel]+PTS_PER_AUDIO_BYTE*audio_chunk_size);
    #endif
    PTSstored[channel] += PTS_PER_AUDIO_BYTE*audio_chunk_size;

    /* Keep track total bytes put here */
    audio_buffer_total[channel]+=audio_chunk_size;
    /* 1st time here? */
    if (audio_buffer[channel]==NULL) {
        audio_buffer[channel] = (gint16*)malloc(AUDIO_BUFFER_SIZE);
    }
    /* Space left in buffer? */
    if (((glong)audio_buffer_index[channel]+(glong)audio_chunk_size) > (AUDIO_BUFFER_SIZE-1)) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio Add buffer overflow, resetting audio buffer");
        #endif
        reset_audio_buffer(channel);
        return;
    }

    /* Need to rotate buffer? */
    if (((glong)audio_buffer_index[channel]+(glong)audio_chunk_size+(glong)audio_buffer_base[channel]) > (AUDIO_BUFFER_SIZE-1)) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Add buffer rotating buffer");
        #endif
        memmove(audio_buffer[channel],(gint16*)((glong)audio_buffer[channel]+(glong)audio_buffer_base[channel]),(glong)audio_buffer_index[channel]);
        audio_buffer_base[channel] = 0;
    }

    /* Copy audio chunk into buffer and update index */
    audio_buffer_dest = (gint16*)((glong)audio_buffer[channel] + (glong)audio_buffer_base[channel] + (glong)audio_buffer_index[channel]);
    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio Add buffer : adding new audio to buffer (dest=%p, source=%p, size=%li)",audio_buffer_dest,audio_chunk,audio_chunk_size);
    #endif
    fast_memcpy(audio_buffer_dest,audio_chunk,(size_t)audio_chunk_size);
    audio_buffer_index[channel] += audio_chunk_size;

    #ifdef DEBUG
    audio_buffer_status(channel);
    #endif

    return;
}


/* Add audio data to output */
gboolean output_audio(glong frame) {
    static gdouble SCR = -1;
    gint16* audio_chunk;
    static gdouble result;
    gint audio_bytes_to_fetch;

    if (demuxing == FALSE) return FALSE;

    if (encode_frame(NULL,NULL,(gdouble)0,FALSE,TRUE) == FALSE) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Output_audio: Nothing in video buffer, canceling audio handling");
        #endif
        return FALSE;
    } else {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: Video buffer has data, continueing audio handling");
        #endif
    }

    /* Init SCR & calc AUDIO_BYTES_PER_FRAME */
    if (SCR==-1) {
        if (PTSaudio!=-1) {
            SCR = (gdouble)PTSaudio;
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: SCR initialised to %f",SCR);
        } else {
            return FALSE;
        }
    }

    /* Calc AUDIO_BYTES_PER_FRAME */
    if (AUDIO_BYTES_PER_FRAME==-1) {
        AUDIO_BYTES_PER_FRAME = ceil(drip_audio_samplerate() / drip_mpeg_framerate()) * 4;
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: AUDIO_BYTES_PER_FRAME = %g",AUDIO_BYTES_PER_FRAME);
    }

    /* Output audio till it's in sync with video */
    /* Amount of audio bytes = rounded UP */
    audio_buffer_filled = FALSE;
    avifile_audio_outstream_bytes_calculated += AUDIO_BYTES_PER_FRAME;
    audio_bytes_to_fetch = (gint)ceil(avifile_audio_outstream_bytes_calculated - (gdouble)avifile_audio_outstream_bytes);
    audio_bytes_to_fetch = (gint)ceil(((gdouble)audio_bytes_to_fetch)/2);
    audio_bytes_to_fetch = audio_bytes_to_fetch << 1;

    result = 0;
    audio_chunk = fetch_from_audio_buffer(audio_bytes_to_fetch,(gint)SCRframe,0);
    if (audio_chunk != NULL) {
        decoding_audio = TRUE;
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: Encoding audio frame with audio_bytes_to_fetch = %li (chunk=%p)",audio_bytes_to_fetch,audio_chunk);
        #endif
        avifile_audio_outstream[0]->AddData(audio_chunk,(gint)audio_bytes_to_fetch/2);
        if (Config.anum == 2) { // 2nd encoded audio channel
            #ifdef DEBUG
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: Fetching audio for 2nd channel");
            #endif
            audio_chunk = fetch_from_audio_buffer(audio_bytes_to_fetch,(gint)SCRframe,1);
            avifile_audio_outstream[1]->AddData(audio_chunk,(gint)audio_bytes_to_fetch/2);
        }
        avifile_audio_outstream_bytes += audio_bytes_to_fetch;
        audio_buffer_filled = TRUE;
        result = encode_frame(stream,im,SCR,TRUE,FALSE);
        if (result != FALSE) {
            SCR += PTS_PER_AUDIO_BYTE*AUDIO_BYTES_PER_FRAME;
        } else {
            #ifdef DEBUG
            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Output_audio: Encoding frame failed..");
            #endif
            return FALSE;
        }
        decoding_audio = FALSE;
    } else {
        /* Got a video frame, but NO audio chunk */
        avifile_audio_outstream_bytes_calculated -= AUDIO_BYTES_PER_FRAME;
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: NO audio chunk available");
        #endif
        return FALSE;
    }

    /* Update, clean & exit */
    return TRUE;
}


/* Handle audio blocks */
void a52_decode_data(guint8 *data, guint8 *end, glong PTS, glong PTSpos, gint channel) {
    /* sample_rate and flags are static because this routine could
     * exit between the a52_syncinfo() and the ao_setup(), and we want
     * to have the same values when we get back ! */
    gint flags;
    sample_t level; // Volume adjustment, TODO less for 5.1?
    sample_t bias;
    gint i;
    gint pcm_size;
    gint len,len_max;
    gint length;

    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"a52_decode_data, channel = %i, PTS = %li, PTSpos = %li",channel,PTS,PTSpos);
    #endif

    start[channel] = data;
    while (1) { 
        len = end - start[channel];
        if (!len)  break; /* Exit decode loop */
        len_max = bufpos[channel] - bufptr[channel];
        if (len > len_max)  len = len_max;
        fast_memcpy(bufptr[channel],start[channel],len);
        bufptr[channel] += len;
        start[channel] += len;

        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"a52 start=%li, data=%li, start-data+1=%li, PTSpos = %li",start[channel],data,(glong)start[channel]-(glong)data,PTSpos);
        #endif
        if ((PTS!=-1) && (((glong)start[channel]-(glong)data)>PTSpos)) {
            /* Only set the PTS once we've hit the magic pointer */
            #ifdef DEBUG
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"a52 PTSa52 set to %i",PTS);
            #endif
            PTSa52[channel] = PTS;
            PTS = -1;
        }

        if (bufptr[channel] == bufpos[channel]) {
            if (bufpos[channel] == &a52_buffer[channel][7]) {
                length = a52_syncinfo(a52_buffer[channel],&flags,&audio_sample_rate,&audio_bit_rate);
                if (!length) {
                    #ifdef DEBUG
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"a52_syncinfo returned 0 lenght");
                    #endif
                    for (bufptr[channel] = a52_buffer[channel];  bufptr[channel] < &a52_buffer[channel][6];  bufptr[channel]++) {
                        bufptr[channel][0] = bufptr[channel][1];
                    }
                    continue;
                }
                bufpos[channel] = &a52_buffer[channel][length];
            } else {
                flags = A52_ADJUST_LEVEL | A52_STEREO;
                level = 1.0; // Volume adjustment, TODO less for 5.1?
                bias = 384.0;
                if (a52_frame(output_liba52[channel],a52_buffer[channel],&flags,(sample_t*)&level,bias)) {
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio decoder: Channel %i, Error in a52_frame",channel);
                    goto error;
                }
                flags &= A52_CHANNEL_MASK;
                audio_channels = 2; // flags[channel] = A52_ADJUST_LEVEL | A52_STEREO
                // Decode audio 
                if (PTSa52[channel]>-1) {
                    for (i=0; i<6; i++) {
                        if (a52_block(output_liba52[channel])) {
                            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio decoder: Channel %i, Error in a52_block",channel);
                            goto error;
                        }
                        pcm_size = float_to_int_stereo(a52_samples(output_liba52[channel]),gint16_samples);
                        #ifdef WRITEOUT_AUDIO
                        write(writeout_audio_fd,gint16_samples,pcm_size);
                        #endif
                        add_to_audio_buffer(gint16_samples,pcm_size,(gint)PTSa52[channel],channel);
                        PTSa52[channel] += PTS_PER_AUDIO_FRAME;
                    }
                } else {
                    #ifdef DEBUG
                    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"a52 block not added, PTS still set to -1");
                    #endif
                }

                bufptr[channel] = &a52_buffer[channel][0];
                bufpos[channel] = &a52_buffer[channel][7];
                continue;
            error:
                bufptr[channel] = &a52_buffer[channel][0];
                bufpos[channel] = &a52_buffer[channel][7];
            }
        }
    }
    return;
}


/* THREAD for reading mpeg2 data ; this way IO and encoding could be parallel if threads supported ?? */
gpointer mpeg2_reader_handler(gpointer data) {
    size_t rbuf,rbuf2;
    gint buffer = 0;
    gint read_size,i;

    while (mpeg2_reader_handler_active == TRUE) { usleep(100);} //poll (...);
    mpeg2_reader_handler_active = TRUE;
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"mpeg2_reader_handler spawned");
    /* Reset buffer */
    for (i=0;i<2;i++) {
        read_buffer[i].lenght = 0;
        read_buffer[i].status = EMPTY;
    }
    mpeg2_reader_handler_pid = getpid();
    pthread_mutex_lock(&mutexmpeg2);
 
    do {
        if (read_buffer[buffer].status != EMPTY && demuxing == TRUE) { 
            pthread_mutex_lock(&mutexmpeg2);
        }
        if (demuxing == FALSE) break;
        rbuf = fread(read_buffer[buffer].buffer, 1, DEMUX_BUFFER_SIZE, in_file);
        read_buffer[buffer].lenght = rbuf;
        
        if (rbuf != ((glong)DEMUX_BUFFER_SIZE) && input_files[input_files_index]!=NULL) {
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Could only read %lu bytes, opening next file |%s|",rbuf,input_files[input_files_index]);
            fclose(in_file);
            in_file = fopen(input_files[input_files_index],"rb");
            if (!in_file) {
                g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"%s - couldnt open file |%s|", strerror(errno),input_files[input_files_index]);
            } else {
                g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Opened %s for input",input_files[input_files_index]);
            }
            input_files_index++;
            read_size = ((glong)DEMUX_BUFFER_SIZE) - rbuf;
            rbuf2 = fread(read_buffer[buffer].buffer+rbuf, 1, read_size, in_file);
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Read left %lu bytes (%u total in buffer)",rbuf2,rbuf+rbuf2);
            rbuf += rbuf2;
            read_buffer[buffer].lenght = rbuf;
        }
        /* Mark status read */
        read_buffer[buffer].status = READ;
        /* flip buffer */ 
        buffer++;  if (buffer>1) { buffer = 0;}
    } while ( demuxing == TRUE ); 
    /* Clean & Close */
    mpeg2_reader_handler_active = FALSE;
    pthread_exit(0);
    return NULL;
}


/* Demux stream */
gpointer drip_demux(gpointer data, gint** audio_channel) {
    static gint mpeg1_skip_table[16] = {
             1, 0xffff,      5,     10, 0xffff, 0xffff, 0xffff, 0xffff,
        0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff
    };
#define DEMUX_HEADER 0
#define DEMUX_DATA 1
#define DEMUX_SKIP 2
    //static gint state = DEMUX_SKIP;
    //static gint state_bytes = 0;
    guint8 *buf;
    size_t rbuf; //,rbuf2;
    guint8 *end = NULL;
    guint8 *tmp1 = NULL;
    guint8 *tmp2;
    gint complain_loudly;
    gint len;
    //gint read_size;
    //guint headerlen, streamid;
    guint packetlen;
    guint a1, a2;
    glong PTS, DTS, SCR;
    //gint pass = 0;
    //glong bufstart = (glong)buf;
    //gint bytes;
    //static uint8_t head_buf[268];
    glong buf_previous = 0;
    gint audio_len;
    gint audio_bytes;
    gint buffer = 0;
    gint buffer_previous = 0;


    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Audio channel 0 = %i, 1 = %i,  Config.anum = %i",*audio_channel[0],*audio_channel[1],Config.anum);
    #endif

    complain_loudly = 1;
    buf = read_buffer[buffer].buffer;

    do {
        while ((read_buffer[buffer].status != READ) && (quit==FALSE)) { usleep(50);} //poll (...);
        if (quit==TRUE) {
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Quitting from inner loop");
            demuxing = FALSE;
            break;
        }
        rbuf = read_buffer[buffer].lenght;
        buf = read_buffer[buffer].buffer;
        read_buffer[buffer].status = PROCESSING;
        end = buf + rbuf;
        //g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: inner loop");
        while (( buf+4 <= end) && (quit == FALSE)) {
            /* check start code */
            buf_previous = (glong)buf;
            if (buf[0] || buf[1] || (buf[2] != 0x01)) {
                if (complain_loudly) {
                    if ((buf[0] == 0) && (buf[1] == 0) && (buf[2] == 0))
                    complain_loudly = 0;
                }
                buf++;
                continue;
            }
            /* Get SCR */
            a1 = buf[4] << 24 | buf[5] << 16 | buf[6]  << 8 | buf[7];
            a2 = buf[8] << 24 | buf[9] << 16 | buf[10] << 8 | buf[11];
            SCR  = (a1 & 0x38000000) << 3;    // shifting the top bit out ..  . ...
            SCR |= (a1 & 0x03fff800) << 4;
            SCR |= (a1 & 0x000003ff) << 5;
            SCR |= (a2 & 0xf8000000) >> 27; 

            /* Get PTS or DTS */
            if( buf[7] & 0x80 ) { /* PTS */
                PTS  = (buf[ 9] & 0x0E) << 29 ;
                PTS |=  buf[10]         << 22 ;
                PTS |= (buf[11] & 0xFE) << 14 ;
                PTS |=  buf[12]         <<  7 ;
                PTS |= (buf[13] & 0xFE) >>  1 ;
            } else {
                PTS = 0;
            }
            if( buf[7] & 0x40 ) { /* DTS */
                DTS  = (buf[14] & 0x0E) << 29 ;
                DTS |=  buf[15]         << 22 ;
                DTS |= (buf[16] & 0xFE) << 14 ;
                DTS |=  buf[17]         <<  7 ;
                DTS |= (buf[18] & 0xFE) >>  1 ;
            } else {
                DTS = 0;
            }

            /* Handle id */
            switch (buf[3]) {
            case 0xb9:  /* program end code */
                demuxing = FALSE;
                return NULL;
            case 0xbb:  /* pss header */
                tmp1 = buf + (buf[4] << 8) + buf[5] + 5;
                buf = tmp1;
                break;
            case 0xba:  /* pack header */
                /* skip */
                if ((buf[4] & 0xc0) == 0x40) {   /* mpeg2 */
                    SCRstream = SCR;
                    tmp1 = buf + 14 + (buf[13] & 7);
                } else if ((buf[4] & 0xf0) == 0x20) {      /* mpeg1 */
                    SCRstream = SCR;  // TODO: SCR mpeg1 diff from mpeg2??
                    tmp1 = buf + 12;
                //} else if (buf + 5 > end) {
                  //  goto copy;
                } else {
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Demuxer: Weird pack header");
                }
                //if (tmp1 > end) {
                //    goto copy;
               // }
                buf = tmp1;
                break;
            case 0xe0:  /* video */
                tmp2 = buf + 6 + (buf[4] << 8) + buf[5];
                //if (tmp2 > end) {
                //    goto copy;
                //}
                if ((buf[6] & 0xc0) == 0x80) {   /* mpeg2 */
                    tmp1 = buf + 9 + buf[8];
                    if (buf[7] & 0x80) {
                        PTSframe = PTS;
                        mpeg2_pts(&mpeg2dec,(glong)PTSframe);
                    }
                } else {  /* mpeg1 */
                    for (tmp1 = buf + 6; *tmp1 == 0xff; tmp1++) {
                        if (tmp1 == buf + 6 + 16) {
                            buf = tmp2;
                            break;
                        }
                    }
                    if ((*tmp1 & 0xc0) == 0x40)
                        tmp1 += 2;
                    tmp1 += mpeg1_skip_table [*tmp1 >> 4];
                    //TODO: calc pts for mpeg1
                    PTSframe = PTS;
                    mpeg2_pts(&mpeg2dec, (glong)PTSframe);
                }
                if (tmp1 < tmp2) {
                    gint num_frames;
                    if (PTSvideo == -1) {
                        PTSvideo = (glong)PTSframe; //Store PTS of 1st frame
                        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: 1st video frame has PTS %i",PTSvideo);
                    }
                    #ifdef DEBUG
                    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Video frame pts = %i",PTS);
                    #endif
                    num_frames = mpeg2_decode_data(&mpeg2dec, tmp1, tmp2);
                }
                buf = tmp2;
                break;
            case 0xbd: /* Private streams (SPU, Audio, others) */
                PTSprivate = PTS; /* store presentation time */
                len = buf[8];
                audio_len = 10 + buf[8] + 3;
                audio_bytes = 6 + (buf[4] << 8) + buf[5] - audio_len;
                tmp1 = (guint8*)((glong)buf + 6 + (buf[4]<<8) + buf[5]);
                spu_track = buf[len+9] & 0x1F;
                audio_track = buf[len+9] & 0x0F;
                #ifdef DEBUG
                g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Private streams, id = %i",buf[len+9]);
                #endif
                // Subpictures
                if ((buf[len+9] & 0xE0) == 0x20) {
                    #ifdef DEBUG
                    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Got a SPU package, pts = %i",PTS);
                    #endif
                    spu_render(buf-14,PTSprivate); /* Call SPU handler of libdripspu */
                } 
                // Audio
                if (((buf[len+9] & 0xF0) == 0x80) && (audio_track==*audio_channel[0] || audio_track==*audio_channel[1])) {
                    if (PTSaudio == -1) {
                        PTSaudio = PTSprivate; //Store PTS of 1st frame
                        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: 1st audio frame has PTS %i",PTSaudio);
                    }
                    /* A52 audio */
                    #ifdef DEBUG
                    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Audio PTS reads %li",PTSprivate);
                    #endif
                    if(audio_channel>=0) {
                        if (audio_track==*audio_channel[0]) {
                            a52_decode_data((guint8*)((glong)buf+len+13),(guint8*)((glong)buf+len+13+audio_bytes),PTSprivate,buf[len+11]<<8|buf[len+12],0);
                        } else if (audio_track==*audio_channel[1]) {
                            a52_decode_data((guint8*)((glong)buf+len+13),(guint8*)((glong)buf+len+13+audio_bytes),PTSprivate,buf[len+11]<<8|buf[len+12],1);
                        }
                    }
                } else if (((buf[len+9] & 0xF0)==0xA0) || (audio_track==(*audio_channel[0]-16)) || (audio_track==(*audio_channel[1]-16))) {
                    /* PCM audio */
                    #ifdef DEBUG
                    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: PCM audio");
                    #endif
                    guint pcm_offset;
                    guint offset = (buf[len+10]<<8) + buf[len+11];
                    const gdouble PTSsub_const = (90000.0/drip_audio_samplerate())/4;
                    gint PTSsub = (gint)((double)(offset)*PTSsub_const);
                    PTSprivate = PTS - PTSsub;
                    packetlen = (buf[4]<<8) + buf[5] - 3 - buf[8];
                    for (pcm_offset=0; ++pcm_offset<packetlen-1; ) {
                        if (buf[len+9+pcm_offset]==0x01 && buf[len+9+pcm_offset+1]==0x80 ) {
                            /* Start */
                            pcm_offset += 2;
                            break;
                        }
                    }
                    if (audio_channel>=0) {
                        guint8* audio_data = (guint8*)((glong)&buf[len+9+pcm_offset]);
                        guint8 tmp;
                        for (guint i=0;i<(packetlen-pcm_offset);i+=2) {
                           tmp = audio_data[i];
                           audio_data[i] = audio_data[i+1];
                           audio_data[i+1] = tmp;
                        }
                        if (PTS) {
                            if (audio_track==*audio_channel[0]) {
                            add_to_audio_buffer((gint16*)(audio_data),packetlen-pcm_offset,PTSprivate,0);
                            } else {
                                add_to_audio_buffer((gint16*)(audio_data),packetlen-pcm_offset,PTSprivate,1);
                            }
                        } else {
                            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Demuxer: no PTS on PCM audio frame");
                        }
                    }
                }
                buf = tmp1;
                break;
            default:
                if (buf[3] < 0xb9) {
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Demuxer: looks like a video stream, not system stream");
                }
                /* skip */
                tmp1 = buf + 6 + (buf[4] << 8) + buf[5];
                //if (tmp1 > end)
                //    goto copy;
                buf = tmp1;
                break;
            }
        }
        //if (buf < end) {
        //copy:
            /* we only pass here for mpeg1 ps streams */
            //g_memmove (demux_buffer, buf, end - buf);
        //}
        read_buffer[buffer].status = EMPTY;
        pthread_mutex_unlock(&mutexmpeg2);
        /* flip buffer */
        buffer_previous = buffer;
        buffer++;  if (buffer>1) { buffer = 0;}
    } while ( (end == read_buffer[buffer_previous].buffer+DEMUX_BUFFER_SIZE) && (quit==FALSE) );
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: end = %p, read_buffer[buffer_previous].buffer+DEMUX_BUFFER_SIZE = %p, quit = %i",end,read_buffer[buffer_previous].buffer+DEMUX_BUFFER_SIZE,quit);
    demuxing = FALSE;
    pthread_mutex_unlock(&mutexmpeg2);

    /* Close libraries */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: Ended task, stopping");
    return NULL;
}


/* Libmpeg2 has a new frame ready */
static void drip_draw_frame(vo_frame_t *frame) {
    total_frames++; // track total frames build
    drip_framerate(&mpeg2dec); // set framerate to current
    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"drip_draw_frame: pts=%u, pts_current=%u, pts_previous=%u",(&mpeg2dec)->pts,(&mpeg2dec)->pts_current,(&mpeg2dec)->pts_previous);
    #endif
    //if (quit==TRUE) return;
    if (processor!=NULL) {
        processor(frame->base,frame->pts); // Call processor [set in drip_mpeg_init()]
    } else {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Cant process frame, this is a bug");
    }
    return;
}


/* Start demuxing dvd stream */
void drip_mpeg_start_demuxing(gint** audio_channel) {
    gint i;
    /* Start (threadless?) demuxing */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Starting demux thread");
    demuxing = TRUE;
    process = TRUE;
    quit = FALSE;
    /* Default audio vari's */
    for (i=0;i<AMAX;i++) {
        audio_buffer[i] = NULL;
        audio_buffer_index[i] = 0;
        audio_buffer_base[i] = 0;
        audio_buffer_total[i] = 0;
        audio_fetched_total[i] = 0;
        IDnumber[i] = 0;
        PTSindex[i] = 0;
        non_silence[i] = FALSE;
        /* A52 vari's */
        bufptr[i] = &a52_buffer[i][0];
        bufpos[i] = &a52_buffer[i][7];
        PTSa52[i] = -1;
        start[i] = 0;
        avifile_audio_total_offset[i] = 0;
        offset_corrected[i] = FALSE;
    }
    drip_demux(NULL,audio_channel);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Ended demux thread");
    return;
}


/* Initialse libvo output */
static gint drip_null_setup(vo_instance_t *instance, gint w, gint h) {
    gint result;
    // publice w & h
    width = w;
    height = h;
    result = libvo_common_alloc_frames(instance,width,height,sizeof(null_frame_t),NULL,NULL,drip_draw_frame);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"drip_null_setup done (%i)",result);
    return result;
}


/* Open libvo output */
vo_instance_t* drip_vo_null_open(void) {
    vo_driver_t *drivers;
    vo_open_t *output_open = NULL;
    gint i;
    null_instance_t* instance;
    const gchar* driver_name = "drip";

    if (strcmp(driver_name,"drip")==0) {
        instance = (null_instance_t*)malloc(sizeof(null_instance_t));
        if (instance==NULL) return NULL;
        instance->vo.setup = drip_null_setup;
        instance->vo.close = libvo_common_free_frames;
        instance->vo.get_frame = libvo_common_get_frame;
        vo_open_instance = (common_instance_t*)instance;
        return (vo_instance_t*)instance;
    } else {
        drivers = vo_drivers();
        for (i=0; drivers[i].name != NULL; i++) {
            printf("driver %s\n",drivers[i].name);
            if (strcmp (drivers[i].name, driver_name) == 0) {
                output_open = drivers[i].open;
            }
        }
        return (vo_instance_t*)output_open;
    }
}


/* Returns NULL audio driver of libao */
ao_open_t *drip_audio_null_driver(void) {
    ao_driver_t *drivers;
    ao_open_t *driver = NULL;
    /* Get drivers list */
    drivers = ao_drivers ();
    for (gint i=0; drivers[i].name!=NULL; i++) {
        if (strcmp(drivers[i].name, "null") == 0) {
             driver = drivers[i].open;
        }
    }
    if (driver==NULL) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Could not open audio output");
    return driver;
}


/* Opens libmpeg2 handlers */
gint drip_mpeg_open(gchar *input,void handler(guint8*[3],uint32_t)) {
    gint i;
    uint32_t accel = 1; 
    // open input file
    demuxing = TRUE;
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Opening input mpeg from %s",input);
    input_files = g_strsplit(input,":",256);
    input_files_index = 0;
    in_file = fopen(input_files[input_files_index],"rb");
    if (!in_file) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"MPEG Open: %s - couldnt open file %s", strerror(errno),input_files[input_files_index]);
    else g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Opened %s for input",input_files[input_files_index]);
    input_files_index++;

    // alloc read buffer
    for (i=0;i<2;i++) {
        if (read_buffer[i].buffer != NULL) free(read_buffer[i].buffer);
        read_buffer[i].buffer = (guint8*)malloc(DEMUX_BUFFER_SIZE);
        read_buffer[i].lenght = 0;
        read_buffer[i].status = EMPTY;
    }
    // init thread mutex
    pthread_mutex_init(&mutexmpeg2, NULL);
    // start reader thread; hope we run with kernel threads so HD reading wont lock encoding
    mpeg2_reader_id = pthread_create(&mpeg2_reader,NULL,mpeg2_reader_handler,NULL);
    // process handler
    processor = handler;
    // open libmpeg2 output
    video_output = vo_open(drip_vo_null_open);
    // init libmpeg2 (video)
    //mpeg2dec = (mpeg2dec_t*)malloc(sizeof(mpeg2dec_t));
    memset(&mpeg2dec,0,sizeof(mpeg2dec_t));
    mpeg2_init(&mpeg2dec,accel,video_output);

    // open liba52 output
    audio_output_open[0] = drip_audio_null_driver();	audio_output_open[1] = drip_audio_null_driver();
    audio_output[0] = ao_open(audio_output_open[0]);	audio_output[1] = ao_open(audio_output_open[1]);
    // init silence chunk
    memset(silence,256*2*sizeof(gint16),0);
    // alloc a52 buffers
    a52_buffer[0] = (guint8*)malloc(3840*sizeof(guint8));
    a52_buffer[1] = (guint8*)malloc(3840*sizeof(guint8));
    // init liba52 (audio)
        #ifdef WRITEOUT_AUDIO
        writeout_audio_fd = open("/tmp/writeout_audio.raw",O_WRONLY|O_CREAT|O_TRUNC,S_IRWXU);
        #endif
    output_liba52[0] = a52_init(accel);			output_liba52[1] = a52_init(accel);
    if (output_liba52==NULL) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Failed to initialize audio");
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Initialized import of mpeg video and audio");
    return TRUE;
}


/* Closes libmpeg2 handlers */
gint drip_mpeg_close(gboolean lock) {
    gint result;
    const char* domain;

    /* Define logging domain (locked\unlocked) */
    if (lock) {
        domain = DRIP_LD;
    } else {
        domain = DRIP_CB_LD;
    }
    g_log(domain,G_LOG_LEVEL_DEBUG,"mpeg_close");
    /* Close mpeg and a52 handlers */
    demuxing = FALSE;

    mpeg2_close(&mpeg2dec);
    vo_close(video_output);
    while (decoding_audio == TRUE) { usleep(20);}
    a52_free(output_liba52[0]);
    a52_free(output_liba52[1]); // Crashes?
    ao_close(audio_output[0]);
    ao_close(audio_output[1]);
    /* Shutdown */
    quit = TRUE;
    //while (demuxing == TRUE) { usleep(53);}
    demuxing = FALSE;
    g_log(domain,G_LOG_LEVEL_DEBUG,"mpeg_close2");
    g_strfreev(input_files);
    reset_audio_buffer(0);				reset_audio_buffer(1);
    fclose(in_file);
        #ifdef WRITEOUT_AUDIO
        close(writeout_audio_fd);
        #endif
    // unlock thread mutex
    pthread_mutex_unlock(&mutexmpeg2);
    /* Cancel mpeg2 reader thread */
    result = pthread_join(mpeg2_reader,NULL);
    // close thread mutex
    pthread_mutex_destroy(&mutexmpeg2);
    return TRUE;

}

